from pathlib import Path
from py_arg_reports.logs import get_logger
from py_arg_reports.reporters.libro_sueldo.data import translate_data
from py_arg_reports.tools.pdf import CanvasPDF, CanvaPDFBlock, Format, Rect
from py_arg_reports.tools.recibos_utils import float_to_format_currency
from reportlab.lib.units import cm


log = get_logger(__name__)
F14 = Format(font_size=14)
F10 = Format(font_size=10)
F9 = Format(font_size=9)
t10_line_sep = 0.5
t9_line_sep = 0.3


def descargar_libro(json_data: dict, output_path: str, filename: str) -> tuple:
    """ Descarga el libro sueldo en formato PDF,
        Retorna una tupla:
         - False, error: si falla
         - True, None: si todo salió bien
    """

    try:
        info_recibo = translate_data(json_data)
    except Exception as e:
        error_detail = f'Error al traducir los datos para libro sueldo: {str(e)}'
        log.error(error_detail)
        return False, error_detail

    # Cada liquidación va a tener su propia carpeta en download
    if not Path(output_path).exists():
        # create the folder
        Path(output_path).mkdir(parents=True, exist_ok=True)
    my_file_path = Path(output_path) / f'{filename}.pdf'
    log.info(f'Generando libro sueldo en {my_file_path}')

    # Mi PDF general donde voy a agregar los bloques
    PDF = CanvasPDF(
        file_path=str(my_file_path),
        title='Libro Sueldo', units=cm,
        data=info_recibo,
    )
    log.info(f'Creando PDF en {my_file_path}')

    draw_header(PDF)
    draw_footer(PDF)

    pos_y = PDF.last_y + 0.1
    empleado_h = 5
    total_rem = 0
    total_no_rem = 0
    total_desc = 0
    for empleado in info_recibo['empleados']:
        log.info(f'Generando empleado {empleado["legajo"]}')
        draw_empleado(PDF, empleado, start_y=pos_y, height=empleado_h)
        pos_y = PDF.last_y + 0.3
        total_rem += empleado['totales_liquidacion']['total_remunerativo']
        total_no_rem += empleado['totales_liquidacion']['total_no_remunerativo']
        total_desc += empleado['totales_liquidacion']['total_retenciones']

    # Agregar bloque de totales
    totales = CanvaPDFBlock(PDF, Rect(0, pos_y, 0, 3), Format(font_size=10, fill_color='#F0F0F099'))
    totales.text('Totales', bold=True, x=1, y=0.5, format_=F14)
    total_empleados = len(info_recibo['legajos'])
    totales.text(f'Cantidad de empleados: {total_empleados}', bold=True, x=12, y=0.5, format_=F10)

    total_rem2 = float_to_format_currency(total_rem)
    totales.text(f'Remunerativos {total_rem2}', bold=True, x=1, y=1.5)
    total_no_rem2 = float_to_format_currency(total_no_rem, 2)
    totales.text(f'No Remunerativos {total_no_rem2}', bold=True, x=7.3, y=1.5)
    total_desc2 = float_to_format_currency(total_desc)
    totales.text(f'Descuentos $ {total_desc2}', bold=True, x=14, y=1.5)

    neto = total_rem + total_no_rem - total_desc
    neto = float_to_format_currency(neto)
    totales.text(f'Neto a cobrar {neto}', bold=True, x=1, y=2.5, format_=F14)

    PDF.finish_page()
    PDF.save()
    return True, None


def draw_header(PDF: CanvasPDF):
    info_recibo = PDF.data
    log.info(f'Creando header en pagina {PDF.page}')
    # Agregar el bloque de headers izquiero
    header = CanvaPDFBlock(PDF, is_header=True, rect=Rect(0, 0, 0, 3.3), format_=Format(font_size=10, fill_color='#D0D0D0CC'))
    header.text('Hojas Móviles Libro Art. 52 Ley 20744', align='center', y=0.7, format_=F14)
    header.text('Folio {page}', x=17.5, y=0.7, format_=F10)
    col = [info_recibo['company_name'], info_recibo['domicilio']]
    header.text_column(col, start_x=0.1, start_y=1.3, format_=Format(font_size=10))
    # actividades
    actividades = ['No tenemos actividades 7777', 'Esta deberia ser la secundaria 1', 'Esta deberia ser la secundaria 2']
    apri = f'Actividad principal: {actividades[0]}'
    asec = [f'Actividad secundaria: {act}' for act in actividades[1:]]
    col = [apri] + asec
    header.text_column(col, start_x=0.4, start_y=2.2, line_sep=t9_line_sep, format_=F9)

    # Agregar el bloque de headers derecho
    per = f'Periodo {info_recibo["tipo_liquidacion"]} {info_recibo["periodo"]}'
    col = ["CUIT: " + info_recibo['cuit'], per]
    header.text_column(col, start_x=13, start_y=1.3, format_=F10)
    return header


def draw_footer(PDF: CanvasPDF):
    log.info(f'Creando footer en pagina {PDF.page}')
    footer = CanvaPDFBlock(
        PDF, is_footer=True,
        rect=Rect(0, PDF.height-1, 0, 0.5), format_=Format(font_size=8, fill_color='#D0D0D0')
    )
    footer.text('Hojas Móviles Libro Art. 52 Ley 20744 -pagina {page}', align='center', y=0.3)
    return footer


def draw_empleado(PDF: CanvasPDF, empleado: dict, start_y, height):
    empleado_block = CanvaPDFBlock(PDF, Rect(0, start_y, 0, height), Format(font_size=10, fill_color='#F0F0F033'))
    name = f'{empleado["legajo"]} - {empleado["nombre"]}'
    y = 0.5
    empleado_block.text(name, bold=True, x=1, y=y)
    lista = ['F. Ingreso', 'F. Egreso', 'Remun. asignada']
    y = y + t9_line_sep + 0.1
    empleado_block.text_column(lista, start_x=1, start_y=y, line_sep=t9_line_sep, format_=F9, bold=True)
    lista = [empleado["fecha_ingreso"], empleado["fecha_ingreso_2"], empleado["basico"]]
    empleado_block.text_column(lista, start_x=4, start_y=y, line_sep=t9_line_sep, format_=F9)
    lista = ['Estado civil', 'Puesto']
    empleado_block.text_column(lista, start_x=7, start_y=y, line_sep=t9_line_sep, format_=F9, bold=True)
    lista = [empleado['estado_civil'], empleado["posicion"]]
    empleado_block.text_column(lista, start_x=10, start_y=y, line_sep=t9_line_sep, format_=F9)
    lista = ['CUIL', 'Categoría', 'Seccion']
    y = 0.5 + t9_line_sep + 0.1
    empleado_block.text_column(lista, start_x=13, start_y=y, line_sep=t9_line_sep, format_=F9, bold=True)
    lista = [empleado["cuil"], empleado["categoria"], empleado["area"]]
    empleado_block.text_column(lista, start_x=16, start_y=y, line_sep=t9_line_sep, format_=F9)

    empleado_block.line(Rect(2, 1.8, 16, 1.8), line_with=2)

    y_titles = 2.3
    # Los tipos de conceptos son:
    # 1 Remunerativos
    # 2 No Remunerativos
    # 3 Descuentos
    conceptos = empleado["conceptos_liquidados"]

    def name_cant(cpt):
        """ No bre y (cantidad) """
        cantidad = cpt.get('cantidad')
        final = cpt['name']
        if cantidad:
            final = f'{final} ({cantidad})'
        return final

    # REMUNERATIVOS

    # Buscar todos los conceptos remunerativos (tipo_concepto=1)
    remunerativos = [c for c in conceptos if c['tipo_concepto'] == 1]
    empleado_block.text('Remunerativos', bold=True, x=0.3, y=y_titles, format_=F10)
    total_remu = sum([cpt['importe'] for cpt in remunerativos])
    empleado_block.text(float_to_format_currency(total_remu), x=5.5, y=y_titles, align='right', bold=True,  format_=F10)

    lista = [name_cant(cpt) for cpt in remunerativos]
    empleado_block.text_column(lista, start_x=0.3, start_y=y_titles + 0.5, line_sep=t9_line_sep, format_=F9)
    lista = [float_to_format_currency(cpt['importe'], include_currency=False) for cpt in remunerativos]
    empleado_block.text_column(lista, start_x=5.5, start_y=y_titles + 0.5, align='right', line_sep=t9_line_sep, format_=F9)

    # NO REMUNERATIVOS

    # Buscar todos los conceptos no remunerativos (tipo_concepto=2)
    no_remunerativos = [c for c in conceptos if c['tipo_concepto'] == 2]
    empleado_block.text('No Remunerativos', bold=True, x=6, y=y_titles, format_=F10)
    total_no_remu = sum([cpt['importe'] for cpt in no_remunerativos])
    empleado_block.text(float_to_format_currency(total_no_remu), x=12, y=y_titles, align='right', bold=True, format_=F10)

    lista = [name_cant(cpt) for cpt in no_remunerativos]
    empleado_block.text_column(lista, start_x=6, start_y=y_titles + 0.5, line_sep=t9_line_sep, format_=F9)
    lista = [float_to_format_currency(cpt['importe'], include_currency=False) for cpt in no_remunerativos]
    empleado_block.text_column(lista, start_x=12, start_y=y_titles + 0.5, align='right', line_sep=t9_line_sep, format_=F9)

    # DESCUENTOS

    # Buscar todos los conceptos descuentos (tipo_concepto=3)
    descuentos = [c for c in conceptos if c['tipo_concepto'] == 3]
    empleado_block.text('Descuentos', bold=True, x=12.4, y=y_titles, format_=F10)
    total_desc = sum([cpt['importe'] for cpt in descuentos])
    empleado_block.text(float_to_format_currency(total_desc), x=18.7, y=y_titles, align='right', bold=True, format_=F10)

    lista = [name_cant(cpt) for cpt in descuentos]
    empleado_block.text_column(lista, start_x=12.4, start_y=y_titles + 0.5, line_sep=t9_line_sep, format_=F9)
    lista = [float_to_format_currency(cpt['importe'], include_currency=False) for cpt in descuentos]
    empleado_block.text_column(lista, start_x=18.7, start_y=y_titles + 0.5, align='right', line_sep=t9_line_sep, format_=F9)

    # NETO

    empleado_block.rectangle(Rect(1, y_titles+1.6, 6, 0.6), fill_color='#D0D0D0AA')
    neto_a_cobrar = float_to_format_currency(empleado["totales_liquidacion"]["neto_liquidacion"])
    empleado_block.text(f'Neto a cobrar   {neto_a_cobrar}', bold=True, x=1.2, y=y_titles+2, format_=F10)


if __name__ == '__main__':
    """
    Esto es para hacer pruebas rapido con los archivos de ejemplo.
    Dejar para hacer correcciones rápidas.
    En estas pruebas escupir los logs en la consola
    """
    import json
    # Descargar el libro
    json_data_test_file = 'py_arg_reports/reporters/libro_sueldo/samples/samples-recibo-info.json'
    json_data = json.load(open(json_data_test_file))
    output_path = 'download/'
    filename = 'libro-sueldo'
    descargar_libro(json_data, output_path, filename)



def fixed_width_str(text, width, align='left', fill_with=' '):
    """
    Returns a string with fixed width. If the text is longer than the width, it will be truncated.
    """
    if not isinstance(text, str):
        text = str(text)
    if len(text) > width:
        return text[:width]
    if align == 'left':
        return text + fill_with * (width - len(text))
    elif align == 'right':
        return fill_with * (width - len(text)) + text
    else:
        raise ValueError(f'Invalid align value: {align}')


def str_dec_num_to_no_dec_sep(str_num):
    """ Tengo un numero como string en un JSON y necesito sus partes entera y decimal """
    importe = float(str_num)
    parte_entera = int(importe)
    parte_decimal = int((importe - parte_entera) * 100)
    return parte_entera, parte_decimal


if __name__ == '__main__':
    assert fixed_width_str('Hello', 10) == 'Hello     '
    assert fixed_width_str('Hello', 10, align='right') == '     Hello'
    assert fixed_width_str('174', 7, fill_with='0') == '1740000'
    assert fixed_width_str('174', 7, fill_with='0', align='right') == '0000174'
    assert fixed_width_str('Hello', 5, align='right') == 'Hello'
    assert fixed_width_str('Hello', 3, fill_with='*') == 'Hel'
    assert fixed_width_str('Hello', 3, align='right', fill_with='*') == 'Hel'



import json
import unittest
from pathlib import Path
import openpyxl
from py_arg_reports.reporters.acreditaciones.galicia import AcreditacionGalicia


class TestAcreditacionGalicia(unittest.TestCase):
    """ Testing para Acreditaciones Santander """

    def setUp(self):
        self.samples_folder = Path('py_arg_reports/reporters/acreditaciones/data')
        self.temp_folder = self.samples_folder / 'temp'
        # Crear la carpeta temporal
        self.temp_folder.mkdir(exist_ok=True)

    def test_acreditacion_galicia_no_data(self):
        data = {}
        with self.assertRaises(ValueError):
            AcreditacionGalicia(data=data)

    def test_acreditacion_galicia_no_nro_cuenta(self):
        file_data = self.samples_folder / 'sample-galicia-missing-nro-cuenta.json'
        data = json.load(open(file_data))
        with self.assertRaises(ValueError) as e:
            AcreditacionGalicia(data)
        expected = 'No hay un nro_cuenta de empleado'
        self.assertIn(expected, str(e.exception))

    def test_acreditacion_galicia_no_pago(self):
        file_data = self.samples_folder / 'sample-galicia-missing-importe-pago.json'
        data = json.load(open(file_data))
        with self.assertRaises(ValueError) as e:
            AcreditacionGalicia(data)
        expected = 'No hay un importe_pago de empleado'
        self.assertIn(expected, str(e.exception))

    def test_acreditacion_galicia_ok(self):
        file_data = self.samples_folder / 'sample-galicia.json'
        data = json.load(open(file_data))
        acreditacion = AcreditacionGalicia(data)
        destination = self.temp_folder / 'galicia.xlsx'
        process, error = acreditacion.generate_file(destination)
        self.assertTrue(process, error)
        # test results excel file
        self.assertTrue(destination.exists())
        # test the content of the excel file
        expected = [
            ['Cuenta', 'Nombre', 'Importe', 'Concepto'],
            ['00209023941002', 'Juan Perez', 1236119.04, 1],
            ['00209023941003', 'Pedro Argento', 5216108.17, 1],
        ]

        wb = openpyxl.load_workbook(destination)
        ws = wb.active
        for row, values in enumerate(expected):
            for col, value in enumerate(values):
                self.assertEqual(ws.cell(row=row+1, column=col+1).value, value)
        wb.close()



import json

from py_arg_reports.reporters.f931.reporter import genera_txt_f931


with open('./test_cases/f931-info.json', 'r') as f:
    f931_data = json.load(f)

cuit_empresa = f931_data['cuit']
periodo = f931_data['periodo']

# Llamo a la función genera_txt_f931
resultado_descarga = genera_txt_f931(
    json_data=f931_data,
    output_path='./test_cases/downloads/',
    filename=f'{cuit_empresa}_{periodo}',
)



from copy import deepcopy
from py_arg_reports.logs import get_logger
from reportlab.lib.colors import HexColor
from reportlab.lib.units import cm
from reportlab.lib.pagesizes import A4
from reportlab.pdfgen.canvas import Canvas


log = get_logger(__name__)


class Rect:
    """ A rectangle, x, y, w, h """
    def __init__(self, x, y, w, h, units=cm):
        self.x = x
        self.y = y
        self.w = w
        self.h = h
        self.units = units

    def __iter__(self):
        """ Para poder pasar los 4 como parametros: *rect """
        return iter([self.x, self.y, self.w, self.h])

    def __str__(self):
        x = round(self.x / self.units, 2)
        y = round(self.y / self.units, 2)
        w = round(self.w / self.units, 2) if self.w else '-'
        h = round(self.h / self.units, 2) if self.h else '-'
        return f'Rect({x}, {y}, {w}, {h})'

    def as_dict(self, units=False):
        x = round(self.x, 2) if not units else round(self.x / self.units, 2)
        y = round(self.y, 2) if not units else round(self.y / self.units, 2)
        w = round(self.w, 2) if not units else round(self.w / self.units, 2)
        h = round(self.h, 2) if not units else round(self.h / self.units, 2)
        return {'x': x, 'y': y, 'w': w, 'h': h, 'units': round(self.units, 2)}

    def copy(self):
        return Rect(self.x, self.y, self.w, self.h, units=self.units)


class Format:
    """ A format for a block """
    def __init__(
            self,
            font='Helvetica', font_bold='Helvetica-Bold',
            font_size=10,
            color='#000000', fill_color='#F0F0F0'
    ):
        self.font = font
        self.font_bold = font_bold
        self.font_size = font_size
        # Los colores son como los de HTML #RRGGBBAA, despues los transformamos al pintar
        self.color = color
        self.fill_color = fill_color

    def __str__(self):
        return f'Format(font={self.font}, font_size={self.font_size}, color={self.color}, fill_color={self.fill_color})'

    def as_dict(self):
        return {
            'font': self.font,
            'font_size': self.font_size,
            'color': self.color,
            'fill_color': self.fill_color,
        }

    def copy(self):
        return Format(
            font=self.font, font_bold=self.font_bold,
            font_size=self.font_size,
            color=self.color, fill_color=self.fill_color
        )


class CanvasPDF:
    """ A PDF file from scratch """
    def __init__(
            self, file_path, title=None,
            pagesize=A4, units=cm,
            margin_x=1, margin_y=1,
            data={},  # datos generales con los que se construye el PDF
    ):
        self.canvas = Canvas(file_path, pagesize=pagesize, bottomup=False)
        self.pagesize = pagesize
        self.units = units
        self.data = data
        self.width_raw, self.height_raw = pagesize
        self.width = self.width_raw / units
        self.height = self.height_raw / units
        self.margin_x = margin_x
        self.margin_y = margin_y
        # Se van agregando bloques y algunos se pasan de la pagina y empieza otra nueva
        # entonces giuardo registro de donde voy para seguir de donde deje
        self.last_y = margin_y
        # llevo registro de la pagina actual
        self.page = 1
        if title:
            # Define a PDF title
            self.canvas.setTitle(title)
        # Define a PDF author
        self.canvas.setAuthor('PayrollT')
        # Define a PDF subject
        self.canvas.setSubject('Software de Liquidación de Sueldos PayrollT')
        # Define a PDF creator
        self.canvas.setCreator('PayrollT creations')
        # Define a PDF producer
        self.canvas.setProducer('PayrollT productions')
        # los bloques de header y de footer se guardan para crear en cada pagina
        self.header_block = None
        self.footer_block = None

    def __str__(self):
        w = round(self.width, 2)
        h = round(self.height, 2)
        mw = round(self.margin_x, 2)
        mh = round(self.margin_y, 2)
        return f'CanvasPDF: ({w} [{mw}] x {h} [{mh}])'

    def finish_page(self):
        """ Terminar la página actual """
        self.canvas.showPage()

    def save(self):
        """ Guardar el PDF """
        self.canvas.save()


class CanvaPDFBlock:
    """ A Canvas PDF block with something, even pages"""

    def __init__(
            self, canvas: CanvasPDF,
            rect: Rect, format_: Format = None,
            is_header=False,  # Si es un header lo guardo
            is_footer=False,  # Si es un footer lo guardo
            is_redraw=False,
            with_rectangles=True
    ):
        """ Inicializamos con coordenadas iniciales y ancho y alto si los hay"""
        # connect to the parent canvas
        # Our PDF canvas
        self.base_pdf = canvas
        self.original_rect = rect.copy() if rect else None
        self.original_format = format_.copy() if format_ else None
        self.history = []  # changes history to re draw
        if is_header:
            self.base_pdf.header_block = self
        if is_footer:
            self.base_pdf.footer_block = self
        self.is_header = is_header
        self.is_footer = is_footer

        # and the "real" canvas
        self.canvas = self.base_pdf.canvas

        # Si este bloque pasa la pagina actual, terminar la pagina
        # al footer lo perdonamos
        ignore_end = is_footer or is_redraw
        if not ignore_end and (rect.y + rect.h > self.base_pdf.height - self.base_pdf.margin_y):
            log.info(f'Block is going to pass the page {self.base_pdf.page}, finishing page')
            self.base_pdf.finish_page()
            # Recomenzar Y
            rect.y = self.base_pdf.margin_y
            self.start_y = rect.y
            self.base_pdf.page += 1
            # re draw header and footer
            if self.base_pdf.header_block:
                log.debug(f'Redrawing HEADER for page {self.base_pdf.page}')
                self.base_pdf.header_block.re_draw()
                rect.y += self.base_pdf.header_block.original_rect.h + 0.1
            if self.base_pdf.footer_block:
                log.debug(f'Redrawing FOOTER for page {self.base_pdf.page}')
                self.base_pdf.footer_block.re_draw()

        # La coordenada Y empieza abajo de todo de la pagina si no ponemos bottomup=False
        # start_x, start_y: coordenadas iniciales para que los numeros internos sean siempre relativos
        # Un cero aqui significa que el bloque está en la esquina superior izquierda de este objeto, no de la página.
        if rect.x < self.base_pdf.margin_x:
            rect.x = self.base_pdf.margin_x
        self.start_x = rect.x
        if rect.y < self.base_pdf.margin_y:
            rect.y = self.base_pdf.margin_y
        self.start_y = rect.y
        # Yo podrìa no saber el ancho y alto, por eso los inicializo en None si es necesario
        max_width = self.base_pdf.width - (self.base_pdf.margin_x * 2)
        if rect.w is None or rect.w == 0 or rect.w > max_width:
            # Default es todo el ancho menos los márgenes
            rect.w = max_width
        elif rect.w < 0:
            # Si quiero un recorte del ancho (por ejemplo la mitad de la página)
            # uso negativos. -2 serìa 1/2
            rect.w = max_width / abs(rect.w)
        self.width = rect.w
        if rect.h is None:
            # Default es todo el alto menos los márgenes
            rect.h = self.base_pdf.height - (self.base_pdf.margin_y * 2)
        self.height = rect.h

        self.format = format_ if format_ else Format()

        # Dibujar rectángulos del bloque
        self.with_rectangles = with_rectangles
        if with_rectangles:
            self.rectangle(
                rect, color=self.format.color,
                fill_color=self.format.fill_color,
                move_start=False
            )
        # Como no se si pase pagina y se creo otra guardo registro del Y maximo donde termine
        # No tener en cuenta si estoy dibujando el footer!
        if not is_footer:
            self.base_pdf.last_y = self.start_y + self.height

    def _rect_to_units(self, rect, move_start=True, move_end=False):
        """ Convierte un rectángulo relativo a uno absoluto """
        # move to start coords
        u = self.base_pdf.units
        x2 = rect.x
        y2 = rect.y
        if move_start:
            # El rectangulo de bloque inicial ya tiene esto considerado
            x2 += self.start_x
            y2 += self.start_y

        x2 = x2 * u
        y2 = y2 * u

        w2 = rect.w or 0
        h2 = rect.h or 0

        if move_end:
            # Esto es una linea, los parametros del final son x2 e y2
            w2 += self.start_x
            h2 += self.start_y

        w2 = w2 * u
        h2 = h2 * u

        # Si bottomup fuera True necesitaria algo asi
        # page_h = self.base_pdf.height * u
        # y3 = page_h - y2
        r = Rect(x2, y2, w2, h2, units=self.base_pdf.units)
        return r

    def rectangle(self, rect: Rect, color=None, fill_color=None, line_width=0.5, rounded=True, move_start=True):
        """ Dibuja un rectángulo en el canvas c """
        log.debug(f'Drawing rectangle at page {self.base_pdf.page}')
        self.history.append(
            {
                'type': 'rectangle',
                'rect': rect.copy(),
                'color': color,
                'fill_color': fill_color,
                'line_width': line_width,
                'rounded': rounded,
                'move_start': move_start,
            }
        )
        # reubicar el rectángulo en el canvas globalself.start_y + rect.y
        rect2 = self._rect_to_units(rect, move_start=move_start)

        color = color if color else self.format.color
        has_alpha = len(color) == 9
        self.canvas.setStrokeColor(HexColor(color, hasAlpha=has_alpha))

        fill_color = fill_color if fill_color else self.format.fill_color
        has_alpha = len(fill_color) == 9
        self.canvas.setFillColor(HexColor(fill_color, hasAlpha=has_alpha))
        self.canvas.setLineWidth(line_width)
        if rounded:
            self.canvas.roundRect(*rect2, radius=3, stroke=1, fill=1)
        else:
            self.canvas.rect(*rect2, stroke=1, fill=1)

    def line(self, rect: Rect, color=None, line_with=1):
        """ Dibujar una linea, en este caso Rect es un punto de inicio y otro de fin """
        log.info(f'Drawing line {rect} at page {self.base_pdf.page}')
        self.history.append(
            {
                'type': 'line',
                'rect': rect.copy(),
                'color': color,
                'line_with': line_with,
            }
        )
        rect2 = self._rect_to_units(rect, move_end=True)
        color = color or self.format.color
        self.canvas.setStrokeColor(HexColor(color))
        # define the border size
        self.canvas.setLineWidth(line_with)
        self.canvas.line(*rect2)

    def text(self, text, x=0, y=0, align='left', bold=False, format_: Format = None):
        """ Dibuja un texto en el canvas c """
        if not text:
            text = ''
        log.debug(f'Drawing text [{text[:10]}] at ({x}, {y}) - page {self.base_pdf.page}')
        if not text:
            return

        self.history.append(
            {
                'type': 'text',
                'text': text,
                'x': x,
                'y': y,
                'align': align,
                'bold': bold,
                'format_': format_,
            }
        )

        # Puede haber valores dinamicos. Por ahora la página
        # Esto podria ser mejor si hay mas valores
        text = text.replace('{page}', str(self.base_pdf.page))

        # Si es centrado y no me da x propongo la mitad del ancho de este bloque
        if align == 'center' and x == 0:
            x = self.width / 2
        # Pasar de coordenadas relativas a absolutas
        u = self.base_pdf.units
        x2 = self.start_x + x
        y2 = self.start_y + y

        x2 = x2 * u
        y2 = y2 * u
        # Si bottomup fuera True necesitaria algo asi
        # y2 = (self.base_pdf.height * u) - y2

        format_ = format_ or self.format
        font = format_.font_bold if bold else self.format.font

        font_size = format_.font_size if format_ else self.format.font_size
        color = format_.color if format_ else self.format.color
        self.canvas.setFont(font, font_size)
        self.canvas.setFillColor(HexColor(color))
        if align == 'left':
            self.canvas.drawString(x2, y2, text)
        elif align == 'right':
            self.canvas.drawRightString(x2, y2, text)
        elif align == 'center':
            # x aqui es el centro del texto
            self.canvas.drawCentredString(x2, y2, text)

    def text_column(self, text_list, line_sep=0.4, start_x=0, start_y=0, align='left', bold=False, format_: Format = None):
        """ Escribir una columna de textos """
        for i, text in enumerate(text_list):
            self.text(text, x=start_x, y=start_y + (line_sep * i), align=align, format_=format_, bold=bold)

    def re_draw(self):
        """ Redibujar el bloque """
        log.info(f'Redraw at page {self.base_pdf.page}')
        block = CanvaPDFBlock(
            canvas=self.base_pdf,
            rect=self.original_rect.copy(), format_=self.original_format.copy(),
            is_header=False, is_footer=False,  # ya lo son
            is_redraw=True,
            with_rectangles=self.with_rectangles,
        )
        # apply history
        history_orig = deepcopy(self.history)
        for h in history_orig:
            type_ = h.pop('type', None)
            log.debug(f'Redrawing {type_} at page {self.base_pdf.page}')
            if type is None:
                err = f'No type in history: {h}'
                log.error(err)
                raise Exception(err)
            if type_ == 'rectangle':
                block.rectangle(**h)
            elif type_ == 'line':
                block.line(**h)
            elif type_ == 'text':
                block.text(**h)



class AcreditacionFile:
    """ Un archivo para un banco (generico)
        El init le permite a las clases que heredan definir muchas cosas.
        Las funciones estan separadas para que cada banco pueda sobreescribir lo que necesite.
    """

    def __init__(
            self, data,
            empresa_required=True,
            liquidacion_required=True,
            include_headers=True,
            empleados_require=['nombre', 'apellido', 'cuil'],
            empresa_require=['razon_social', 'cuit'],
            liquidacion_require=[],
            extras_require=[],
    ):
        self.data = data

        self.empresa = None
        self.empleados = []
        # Datos de esta liquidacion
        self.liquidacion = None
        # Extras que cada banco puede requerir
        self.extras = {}
        # Incluye los headers el archivo final?
        self.include_headers = include_headers

        # Definir que es obligatorio, cada clase lo elige
        # Empresa es requerida? Por ejemplo para Galicia no es requerida
        self.empresa_required = empresa_required
        self.liquidacion_required = liquidacion_required

        # Cada objeto puede definir que campos son requeridos
        self.empleados_require = empleados_require
        self.empresa_require = empresa_require
        self.liquidacion_require = liquidacion_require
        self.extras_require = extras_require

        self.load_data()

    def load_data(self):
        """ Cargar los datos """
        self.load_empleados(require=self.empleados_require)
        self.load_empresa(require=self.empresa_require)
        self.load_liquidacion(require=self.liquidacion_require)
        self.load_extras(require=self.extras_require)

    def load_empleados(self, require=[]):
        """ Cargar los datos de los empleados """
        # validar que tenemos una lista de empleados
        empleados = self.data.get('empleados')
        if not isinstance(empleados, list):
            raise ValueError('No hay una lista de empleados')
        for empleado in empleados:
            empleado = self.validate_empleado(empleado, require)
            self.empleados.append(empleado)

    def validate_empleado(self, empleado, require=[]):
        """ Validar los datos de un empleado """
        for field in require:
            if not empleado.get(field):
                raise ValueError(f'No hay un {field} de empleado')
        return empleado

    def load_empresa(self, require=[]):
        """ Cargar los datos de la empresa """
        empresa = self.data.get('empresa')
        if self.empresa_required and not isinstance(empresa, dict):
            raise ValueError('No hay un diccionario con los datos de la empresa')
        self.empresa = self.validar_empresa(empresa, require) if empresa else None

    def load_liquidacion(self, require=[]):
        """ Cargar los datos de la liquidacion """
        liquidacion = self.data.get('liquidacion')
        self.validar_liquidacion(liquidacion, require)
        self.liquidacion = liquidacion

    def validar_liquidacion(self, liquidacion, require=[]):
        """ Si hay una, validar la liquidacion """
        if not liquidacion:
            if self.liquidacion_required and not isinstance(liquidacion, dict):
                raise ValueError('No hay un diccionario con los datos de la liquidacion')
            self.liquidacion = None
            return

        # validar el total de pagos (si existe)
        if liquidacion.get('total_pago'):
            self.validar_total_pago(liquidacion)

        # ver si hay campos requeridos
        for field in require:
            if not liquidacion.get(field):
                raise ValueError(f'No hay un {field} de liquidacion')

        self.liquidacion = liquidacion

    def validar_total_pago(self, liquidacion):
        """ Ver que el total de pagos sea el mismo a la suma de los pagos de los empleados """
        total_from_empleados = sum(
            [
                float(empleado.get('importe_pago'))
                for empleado in self.empleados
            ]
        )
        # Si total_pago es "0" entonces lo calculo solo sin validar
        if liquidacion.get('total_pago') == "0":
            liquidacion['total_pago'] = str(total_from_empleados)
        else:
            total_pago = float(liquidacion.get('total_pago'))
            if abs(total_from_empleados - total_pago) > 5:
                error = (
                    'El total de pagos no coincide con la suma de los pagos de los empleados'
                    f'{total_from_empleados} != {total_pago}'
                )
                raise ValueError(error)

    def validar_empresa(self, empresa, require=[]):
        """ Validar los datos de la empresa """
        for field in require:
            if not empresa.get(field):
                raise ValueError(f'No hay un {field} de empresa')
        return empresa

    def load_extras(self, require=[]):
        """ Cargar los extras """
        self.extras = self.data.get('extras', {})
        for field in require:
            if not self.extras.get(field):
                raise ValueError(f'No hay un {field} de extras: {self.extras}')

    def generate_file(self, path):
        """ Generar el archivo """
        raise NotImplementedError


class AcreditacionesHeadDetailTrailerFile(AcreditacionFile):
    """Este es un archivo de texto con
        Header  -> generate_header
        Detalle -> generate_detalle_empleado
        Trailer -> generate_trailer
    """

    def generate_file(self, path):
        """ Generar el archivo
            Devuelve un booleano y un mensaje de error (si hay)
        """
        f = open(path, 'w')
        header = self.generate_header()
        f.write(header)
        f.write('\n')
        for empleado in self.empleados:
            detalle_empleado = self.generate_detalle_empleado(empleado)
            f.write(detalle_empleado)
            f.write('\n')
        trailer = self.generate_trailer()
        f.write(trailer)
        f.close()
        return True, None
